"""Script for analysing semgrep results."""
import os
import logging
import argparse
import json
from sarif import loader as sarif_loader  # type: ignore
from jinja2 import Environment, FileSystemLoader
import csv
from typing import Dict, Any, List, Tuple
from analyser import highlight
from secrets import token_hex
import yaml
import csv

logging.basicConfig(
    # filename=f"analyser.log",
    format="%(asctime)s.%(msecs)d %(levelname)-8s"
    + " [%(filename)s:%(lineno)d] %(message)s",
    datefmt="%Y-%m-%d:%H:%M:%S",
    level=logging.DEBUG,
)

parser = argparse.ArgumentParser(
    description="Semgrep result analyser"
)

parser.add_argument(
    "-r",
    "--report-dir",
    help="Directory containing SARIF files from semgrep scan",
    required=True,
    default="/dev/null",
    type=str,
)

parser.add_argument(
    "-c",
    "--csv-file",
    help="CSV file with tuples (dir, repo, branch) for git clones",
    required=False,
    default="/dev/null",
    type=str,
)

parser.add_argument(
    "-f",
    "--filter-file",
    help="SARIF filter",
    required=False,
    default="/dev/null",
    type=str,
)

parser.add_argument(
    "-m",
    "--service-mapping",
    help="Service mapping file",
    required=False,
    default="/dev/null",
    type=str,
)

args = vars(parser.parse_args())

REPORT_DIR: str = str(args.get("report_dir"))
FILTER_FILE: str = str(args.get("filter_file"))
GIT_CSV_FILE: str = str(args.get("csv_file"))
SERVICE_MAPPING_FILE: str = str(args.get("service_mapping"))

rtg: Dict[str, str] = {}
git_mapping: Dict[str, Dict[str, str]] = {}
service_mapping: List[Dict[str, Any]] = []

j2_env = Environment(loader=FileSystemLoader("."))


def metadata(sarif: Any) -> Tuple[str, str, str, str]:
    """
    Return metadata for a sarif file.

    (mapping_group, git_url, git_branch, git_commit)
    """
    git_dir: str = sarif.get_file_name().replace(".json", "")
    git_url: str = git_mapping.get(git_dir, {}).get("url", git_dir)
    git_branch: str = git_mapping.get(git_dir, {}).get("branch", "")
    git_commit: str = git_mapping.get(git_dir, {}).get("commit", "")

    group = rtg.get(git_url, "unassigned")

    return (group, git_url, git_branch, git_commit)


def create_report_for_group(group: str = "", sarifs: List[Any] = []) -> None:
    """Create reports for group of SARIF data."""
    create_report_for_group_html(group, sarifs)
    create_report_for_group_csv(group, sarifs)


def create_report_for_group_html(
  group: str = "",
  sarifs: List[Any] = []
  ) -> None:
    """Create html report for group of SARIF data."""
    html_string = j2_env.get_template("scan_report.jinja2.html").render(
        sarifs=sarifs,
        f_metadata=metadata,
        f_random=token_hex,
        f_highlight=highlight,
    )

    report_file_name = f"{group.replace(' ', '_')}.html"
    with open(f"{REPORT_DIR}/{report_file_name}", "w") as f:
        print(html_string, file=f)
        logging.info(f"Created HTML report: {REPORT_DIR}/{report_file_name}")


def create_report_for_group_csv(
  group: str = "",
  sarifs: List[Any] = []
  ) -> None:
    """Create CSV report for group of SARIF data."""
    report_file_name = f"{group.replace(' ', '_')}.csv"
    csv_file = open(f"{REPORT_DIR}/{report_file_name}", "w")
    csv_writer = csv.writer(csv_file)

    # CSV Header
    csv_writer.writerow(
        [
            "Engineering Group",
            "Repository",
            "Location",
            "Semgrep Rule",
            "Valid",
            "Severity",
            "Comment",
            "Jira Ticket"
        ]
    )

    for sarif in sarifs:
        group_name = metadata(sarif)[0]
        url_repo = metadata(sarif)[1]
        branch = metadata(sarif)[2]
        commit = metadata(sarif)[3]

        location = branch
        if len(commit) > 0:
            location = commit

        for r in sarif.get_results():
            r_ph = r["locations"][0]["physicalLocation"]
            file = "/".join(r_ph["artifactLocation"]["uri"].split("/")[3:])
            start_line = r_ph["region"]["startLine"]
            end_line = r_ph["region"]["endLine"]
            start = r_ph["region"]["startColumn"]
            end = r_ph["region"]["endColumn"]

            url_file = f"{url_repo}/blob/{location}" +\
                f"/{file}#L{start_line}-{end_line}"
            rule = r["ruleId"]

            # Check corresponding items in the CSV header above
            csv_writer.writerow(
                [
                    group_name,
                    url_repo,
                    url_file,
                    rule,
                    "",
                    "",
                    "",
                    ""
                ]
            )

    logging.info(f"Created CSV report: {REPORT_DIR}/{report_file_name}")


def repo_to_group(
        service_mapping: List[Dict[str, Any]] = []
      ) -> Dict[str, str]:
    """
    Return repo to group dict.

    E.g. {"https://gitlab.bit9.local/cb-defense/analytics": "TAU"}.
    """
    rtg = {}

    for entry in service_mapping:
        eng_group = entry.get("Engineering Group", "")
        repos = entry.get("Assets", {}).get("Code Repositories", [])

        for repo in repos:
            rtg[repo] = eng_group

    return rtg


def group_to_sarifs(
        sarifs: List[Any] = [],
        rtg: Dict[str, str] = {}
      ) -> Dict[str, List[Any]]:
    """Return dict eng. group to list of corresponding sarifs."""
    gts: Dict[str, List[Any]] = {}

    gts["all"] = []
    for sarif in sarifs:
        git_dir = sarif.get_file_name().replace(".json", "")
        git_url = git_mapping.get(git_dir, {}).get("url", git_dir)
        group = rtg.get(git_url, "unassigned")
        if group != "unassigned":
            # add "group-" prefix to make report file names easier to read
            group = f"group-{group}"

        gts[group] = gts.get(group, []) + [sarif]
        gts["all"] = gts.get("all", []) + [sarif]

    return gts


if __name__ == "__main__":
    logging.info("Starting semgrep result analyser")

    logging.info(f"Loading sarif files from : {REPORT_DIR}")
    try:
        sarifs = sarif_loader.load_sarif_files(f"{REPORT_DIR}/*.json")
        logging.info(f"Loading successful")
    except Exception as e:
        semgrep_sarif = None
        logging.error(f"Loading failed:\n{e}")

    logging.info("Loading git CSV file")
    with open(GIT_CSV_FILE, "r") as csv_file:
        csv_data = csv.reader(csv_file)

        for row in csv_data:
            try:
                git_mapping[row[0]] = {
                    "url": row[1],
                    "branch": row[2],
                    "commit": row[3]
                }
            except Exception as e:
                pass

    logging.info("Done loading git CSV file")

    if FILTER_FILE != "/dev/null":
        logging.info(f"Loading SARIF filter file: {FILTER_FILE}")
        with open(FILTER_FILE, "r") as filter_file:
            filter_data = yaml.safe_load(filter_file)

            # Apply filter
            sarifs.init_general_filter(
                configuration=filter_data.get("configuration", []),
                filter_description=filter_data.get("description", []),
                include_filters=filter_data.get("include", []),
                exclude_filters=filter_data.get("exclude", [])
            )

        logging.info("Done loading SARIF filter file")

    logging.info(f"Loading git service mapping: {SERVICE_MAPPING_FILE}")
    with open(SERVICE_MAPPING_FILE, "r") as service_mapping_file:
        try:
            data = yaml.safe_load(service_mapping_file)
            service_mapping = data.get("Mapping", [])
        except Exception as e:
            logging.warning(f"Error: {e}")
            service_mapping = []

    logging.info("Done loading service mapping file")

    rtg = repo_to_group(service_mapping)
    gts = group_to_sarifs(sarifs, rtg)

    for group in gts:
        create_report_for_group(group, gts[group])
